#include "quaternion.h"

#include "matrix.h"

#include <math.h>

const Quaternion i(0, 1, 0, 0);   // Imaginary unit
const Quaternion j(0, 0, 1, 0);
const Quaternion k(0, 0, 0, 1);

Quaternion::Quaternion(const Quaternion &Q)
{
	w = Q.w;
   x = Q.x;
   y = Q.y;
   z = Q.z;
}

Quaternion::Quaternion(float s)
{
	w = s;
	v = Eo;
}

Quaternion::Quaternion(const vector &V)
{
	w = 0;
	v = V;
}

Quaternion::Quaternion(float s, const vector &V)
{
	w = s;
	v = V;
}

Quaternion::Quaternion(const matrix &A)
{
	w = 0.5F * sqrtf(tr(A));
	x = (A(3, 2) - A(2, 3)) / (4 * w);
	y = (A(1, 3) - A(2, 1)) / (4 * w);
	z = (A(2, 1) - A(1, 2)) / (4 * w);
}

Quaternion::Quaternion(float Qw, float Qx, float Qy, float Qz)
{
	w = Qw;
	x = Qx;
	y = Qy;
	z = Qz;
}

Quaternion::operator float()
{
	return w;
}

Quaternion::operator vector()
{
	return v;
}

Quaternion::operator matrix()
{
	return matrix(1 - 2 * (y * y + z * z), 2 * (x * y - w * z),     2 * (x * z + w * y),
	              2 * (x * y + w * z),     1 - 2 * (x * x + z * z), 2 * (y * z - w * x),
	              2 * (x * z - w * y),     2 * (y * z + w * x),     1 - 2 * (x * x + y * y));
}

Quaternion Quaternion::operator+() const
{
   return *this;
}

Quaternion Quaternion::operator-() const
{
   return Quaternion(-w, -x, -y, -z);
}

Quaternion Quaternion::operator!() const
{
	float N = w*w + x*x + y*y + z*z;

	return Quaternion(w / N, -x / N, -y / N, -z / N);
}

Quaternion Quaternion::operator~() const
{
	return Quaternion(w, -x, -y, -z);
}

Quaternion &Quaternion::operator=(const Quaternion &Q)
{
	w = Q.w;
   x = Q.x;
   y = Q.y;
   z = Q.z;

   return *this;
}

Quaternion &Quaternion::operator+=(const Quaternion &Q)
{
	w  += Q.w;
	x  += Q.x;
	y  += Q.y;
	z  += Q.z;

	return *this;
}

Quaternion &Quaternion::operator-=(const Quaternion &Q)
{
	w  -= Q.w;
	x  -= Q.x;
	y  -= Q.y;
	z  -= Q.z;

	return *this;
}

Quaternion &Quaternion::operator*=(float r)
{
	w *= r;
	x *= r;
	y *= r;
	z *= r;

	return *this;
}

Quaternion &Quaternion::operator*=(const Quaternion &Q)
{
	float Qw = w * Q.w - x * Q.x - y * Q.y - z * Q.z;
	float Qx = w * Q.x + x * Q.w + y * Q.z - z * Q.y;
	float Qy = w * Q.y - x * Q.z + y * Q.w + z * Q.x;
	float Qz = w * Q.z + x * Q.y - y * Q.x + z * Q.w;

	w = Qw;
	x = Qx;
	y = Qy;
	z = Qz;

	return (*this);
}

Quaternion &Quaternion::operator/=(float r)
{
	w /= r;
	x /= r;
	y /= r;
	z /= r;

	return *this;
}

Quaternion &Quaternion::operator/=(const Quaternion &Q)
{
	return *this *= !(*this);
}

bool operator==(const Quaternion &Q1, const Quaternion &Q2)
{
   if (Q1.w == Q2.w && Q1.x == Q2.x && Q1.y == Q2.y && Q1.z == Q2.z) return true;
   else return false;
}

bool operator!=(const Quaternion &Q1, const Quaternion &Q2)
{
	if(Q1.w != Q2.w || Q1.x != Q2.x || Q1.y != Q2.y || Q1.z != Q2.z)
      return true;
   else
      return false;
}

Quaternion operator+(const Quaternion &Q1, const Quaternion &Q2)
{
	return Quaternion(Q1.w + Q2.w, Q1.x + Q2.x, Q1.y + Q2.y, Q1.z + Q2.z);
}

Quaternion operator-(const Quaternion &Q1, const Quaternion &Q2)
{
	return Quaternion(Q1.w - Q2.w, Q1.x - Q2.x, Q1.y - Q2.y, Q1.z - Q2.z);
}

Quaternion operator*(const Quaternion &Q1, const Quaternion &Q2)
{
	float w = Q1.w * Q2.w - Q1.x * Q2.x - Q1.y * Q2.y - Q1.z * Q2.z;
	float x = Q1.w * Q2.x + Q1.x * Q2.w + Q1.y * Q2.z - Q1.z * Q2.y;
	float y = Q1.w * Q2.y - Q1.x * Q2.z + Q1.y * Q2.w + Q1.z * Q2.x;
	float z = Q1.w * Q2.z + Q1.x * Q2.y - Q1.y * Q2.x + Q1.z * Q2.w;

	return Quaternion(w, x, y, z);
}

Quaternion operator*(float r, const Quaternion &Q)
{
	return Quaternion(r * Q.w, r * Q.x, r * Q.y, r * Q.z);
}

Quaternion operator*(const Quaternion &Q, float r)
{
	return Quaternion(Q.w * r, Q.x * r, Q.y * r, Q.z * r);
}

Quaternion operator/(const Quaternion &Q, float r)
{
	return Quaternion(Q.w / r, Q.x / r, Q.y / r, Q.z / r);
}

Quaternion operator/(const Quaternion &Q1, const Quaternion &Q2)
{
	return Q1 * !Q2;
}

float N(const Quaternion &Q)
{
	return Q.w*Q.w + Q.x*Q.x + Q.y*Q.y + Q.z*Q.z;
}

Quaternion &Quaternion::normalize()
{
   return *this /= sqrtf(N(*this));
}

void Quaternion::from_axisangle(vector &axis,double angle)
{
	double angleRad = DEGTORAD(angle/2.0);
	double sin_a = sin(angleRad);
	double cos_a = cos(angleRad);

	x = axis.x*sin_a;
	y = axis.y*sin_a;
	z = axis.z*sin_a;
	w = cos_a;
	normalize();
}

void Quaternion::from_euler(Point3D<double> &axis)
{
	vector vx(1,0,0),vy(0,1,0),vz(0,0,1);
	Quaternion qx, qy, qz, qt;

	qx.from_axisangle(vx,axis.x);
	qy.from_axisangle(vy,axis.y);
	qz.from_axisangle(vz,axis.z);

	qt=qx*qy;
	*this=qt*qz;
}

Point3D<double> Quaternion::to_euler()
{
	Point3D<double> axis;
    double sqw = w*w;    
	double sqx = x*x;    
	double sqy = y*y;    
	double sqz = z*z;    
	axis.z = RADTODEG(atan2(2.0*(x*y+z*w),(sqx-sqy-sqz+sqw)));  //heading : tete
	axis.x = RADTODEG(atan2(2.0*(y*z+x*w),(-sqx-sqy+sqz+sqw))); //bank: inclinaison
	axis.y = RADTODEG(asin(-2.0*(x*z-y*w)));					//attitude: attitude	
	return axis;
}

/*
void Quaternion::to_euler(Point3D<double> *axis,double *angle)
{
normalize();
double cos_a =w;
double angle = acos(cos_a)*2;
sin_a = sqrt(1.0-cos_a*cos_a);
if (fabs(sin_a)<0.0005) sin_a = 1;
axis.x = x/sin_a;
axis.y = y/sin_a;
axis.z = z/sin_a;
}
*/