/* ***********************************************************************************
	Writer:		Sebastien Bloc
	Copyright:	2003-2004
	eMail:		sebastien.bloc@free.fr
	URL:		http://mignonsoft.free.fr

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	http://www.gnu.org/copyleft/gpl.html
**************************************************************************************
	
/*! \file input.h
\brief Parametre de gestion d'un evenement d'entre (souris + clavier + timer)
*/

#ifndef INCLUDE_INPUT_H
#define INCLUDE_INPUT_H

#include "trigo.h"
#include "myString.h"
#include "mylist.h"
#include "sigslot.h"

/*!
\brief Etat de la souris

Parametre:
	- position (pos)
	- {left,mid,righ}Btn 
	- roulette 
		- wheelVal position de la molette)
		- wheelDelta dernier sens de rotation de la molette
*/
class StateMouse
{
public:
	Point2D<int> pos;
	BOOL leftBtn;
	BOOL rightBtn;
	BOOL midBtn;
	int wheelVal;
	int wheelDelta;

	StateMouse();
	void operator= (StateMouse &src);
	int GetNbButtonPush();
	StateMouse * GetDelta(StateMouse *second); // delta = second-this

	void ToString(MyString &str);
};

#define DebugMouse(context,ptrMouse) { MyString message; message<<context<<" "; if (ptrMouse) (ptrMouse)->ToString(message); else message<<" (NULL)"; Debug0Param(message); }
#define DebugMouse2(context,ptrMouse,face) { MyString message; message<<context<<" "; if (ptrMouse) (ptrMouse)->ToString(message); else message<<" (NULL)"; message<<" Face:"<<face; Debug0Param(message); }

/*!
\brief Parametres des touches en cours
*/
class StateKey
{
private:
	int oldDelay;		//! Encienne valeur de delay du clavier
	int oldRepeate;		//! Envienne valeur de repeate du clavier

public:
	char letter;
	short code;
	BOOL isPush;

	StateKey();
	~StateKey();

	void operator= (StateKey &src);
	void ToString(MyString &str);

	// tools
	static BOOL GetState(int code);
	static void SetDelayAndRepeate(int delay,int repeate);
	void SaveDelayAndRepeate();
	void RestoreDelayAndRepeate();
};

#define DebugKey(context,ptrKey) { MyString message; message<<context<<" "; if (ptrKey) (ptrKey)->ToString(message); else message<<" (NULL)"; Debug0Param(message); }
#define DebugKey2(context,ptrKey,face) { MyString message; message<<context<<" "; if (ptrKey) (ptrKey)->ToString(message); else message<<" (NULL)"; message<<" Face:"<<face; Debug0Param(message); }


/*!
\brief
Gestion de timer avec lancement de callBack

Utilsation:
\code
class ToolTip
{
	Timer onTimeOut;
	ToolTip ()
	{
		BindTimer(onTimeOut,OnTimeOut,10000);
	}
	
	void OnTimeOut(Timer *event)
	{
		// do something ...
		event->Kill(); // si on ne veut executer une seul fois
	}
};

//Traitement dans la callback
LRESULT WINAPI MainProc (HWND hDlg,UINT message,WPARAM wParam,LPARAM lParam)
{
	case WM_CREATE:
		TimerEvent::Init(hDlg);
		break;
	case WM_TIMER:
		TimerEvent::Call(wParam);
		break;
}
\endcode
*/

using namespace sigslot;  // pour signal (evit l'ecriture sigslot::signal?<...>)

/*!
\brief Macro qui genere l'attachement a l'event du timer et la lance
*/
#define BindTimer(theTimer,methode,frec)\
{\
	using namespace sigslot; \
	theTimer.onTimer.connect(this,methode);\
	theTimer.Start(frec);\
}
	
class Timer
{
public:
	// variable membre static
	static MyList<Timer *> list;	//!< list des instances de TimerEvent
	static int nextFreeId;				//!< prochain numero de timer valide
	static HWND hWnd;					//!< hWnd qui collect les messages
	static LARGE_INTEGER frecTimer;		//!< frequence du timer

public:
	// variable membre
	signal1<Timer *> onTimer;			//!< int Call(Timer *timer)
	int id;								//!< id du WM_TIMER en cour
	BOOL insideList;					//!< cette element est il dans la liste ?
	int frequency;						//!< frequence de declenchement
	LARGE_INTEGER beginTime;
	BOOL isRuning;						//!< timer en cours

	static BOOL Init(HWND _hWnd);		// initialise avec la callback gerant les timer
	static Timer *Find(int idToFind);	// recherche d'un instanciation	
	static BOOL Call(int idToFind);		// A utiliser dans le message WM_TIMER de la callback de hWnd

public:
	Timer();
	~Timer();
	BOOL Start();
	BOOL Start(int frequency);
	BOOL Pause();
	BOOL Resume();
	BOOL Stop();
	BOOL StopAndUnBind();
	BOOL Stop(int idToFind,BOOL unBind);

	double GetDeltaTime(BOOL reset=FALSE);
	double GetDeltaTime(BOOL reset,LARGE_INTEGER *referenceTime);

	BOOL IsRuning();
};

#endif