/* ***********************************************************************************
	Writer:		Sebastien Bloc
	Copyright:	2003-2004
	eMail:		sebastien.bloc@free.fr
	URL:		http://mignonsoft.free.fr

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	http://www.gnu.org/copyleft/gpl.html
*********************************************************************************** */

/*! \file face.h
\brief Gestion de Text en 3D
*/

#ifndef INCLUDE_TEXT_3D_H
#define INCLUDE_TEXT_3D_H

#pragma warning( disable : 4217 )
#pragma warning( disable : 4099 )

#include "material.h"
#include <ftgl/ftfont.h>

typedef enum
{
	align_begin,		//!< alignement a gauche si dans les X, en haut si dans les Y
	align_center,		//!< centrage
	align_end,			//!< alignement a droite si dans les X, en bas si dans les Y
	align_justify		//!< mode d'extenssion maximum
} TextAlign;

//! Element de parsing
class ParsingElem
{
public:
	GLfloat deltaX;					//!< taille X du mots
	BOOL isWord;					//!< word ou separator?

	ParsingElem();
	void operator=(ParsingElem &src);
};

//! parsing de mots
class ParsingWord : public ParsingElem
{
public:
	MyString text;

	ParsingWord();
	void operator=(ParsingWord &src);
};

/*!
\brief Sauvegard du parsing du text

parametre:
	- deltaX: taille
	- indBegin,indEnd: indice dans la chaine de debut et de fin de l'element
	- align: type d'alignement
	- words: lsit des pointeurs sur les mots contenu dans la ligne (si l'on est dans une ligne)
	- separator: separateur de l'element (pour un mots: '\t', ' ' etc ...)
*/

class ParsingLine : public ParsingElem
{
public:
	GLfloat deltaY;					//!< taille Y du mots (height de la fonte)
	TextAlign align;				//!< type d'allignement horizontale
	MyList<ParsingWord *> words;	//!< list des pointeur sur les mots de la ligne
	double xJustify;				//!< taille de chaque justification dans les X
	double xPos;						//!< postion x de depart
	int nbChar;						//!< nombre total de caractere

	ParsingLine();
	void operator=(ParsingLine &src);
};

typedef enum
{
	ft_texture,		
	ft_extrude,		
	ft_outline,	
	ft_polygon,		
	ft_bitmap,	
	ft_pixmap,
	ft_unknown
} FontType; 

/*!
\brief Classe de gestion de text 3D (base sur la FTGL)

Les parametres:
	- value: text
	- clipping: taille de la limite d'affichage
	- align[X,Y]: type d'allignement du text par rapport au clipping
	- type: type de fonte (texture,extrude,outline etc ...)
	- font: la fonte qui va etre utilise
	- material: couleur a afficher
	- pos: decalade par rapport a la face

ex.:
\code
BasicText3D text;
FTGLTextureFont font;
font.open("myfont.ttf");
font.FaceSize(18);
text.Setup(&font,ft_texture,0,0);
text.value="hello word !!";
// positionage en 3D par les glTranlate,glRotate,glScall
text.Draw();
\endcode
*/

class BasicText3D 
{
public:
	FontType type;				//!< type de fonte
	FTFont *font;				//!< fonte a appliquer
	Material material;			//!< couleur
	float depth;				//!< profondeur dans le cas d'une font extrude
	int xCenter;				//!< centrage en X (-1 a gauche, 0 centrer, 1 a droite)
	int yCenter;				//!< centrage en Y (-1 en haut,0 centrer, 1 en base)
	GLdouble zCenter;			//!< centrage en Z
	double scale;				//!< taille relative de la fonte (en %)
	BOOL usePicking;			//!< dois on utiliser le picking
	double ratioXY;				//!< rapport X/Y de la surface en cours
	double ratioYX;				//!< rapport Y/X de la surface en cours

public:
	BasicText3D();
	void SetRatioXY(GLdouble value);
	void Setup(FTFont *f,FontType t,int x=0,int y=0);
	void operator=(BasicText3D &src);
	BOOL DrawPicking(char *text);
	BOOL Draw(char *text);
	BOOL Draw(char letter);

	Point3D<float> GetBox(char *text);
	float GetAscender();
	float GetDescender();
	float GetHeight();
	float GetWidth(char *text);
	float GetWidth(char letter);
};

/*!
\brief Classe de gestion d'alignement de text 3D (base sur la FTGL)

Les parametres:
	- clipping: taille de la limite d'affichage (si enableClipping activ)
	- align[X,Y]: type d'allignement du text par rapport au clipping
	- wrapping[line,word,letter]: type de decoupage secable

Example:
\code
Text3D textHautGauche,textCentreBas,textJustifyCentre;
// affiche en du text en haut a gauche du contenant sans justifiaction
textHautGauche.Setup(fontTexture,ft_texture);
textHautGauche.SetClipping((100,100);
textHautGauche.SetAlign(align_begin,align_begin);
textHautGauche.value="Haut a gauche";

// affiche en du text en centr en bas sans justifiaction
textCentreBas.Setup(fontTexture,ft_texture);
textCentreBas.SetClipping((100,100);
textCentreBas.SetAlign(align_center,align_end);
textCentreBas.value="Centr en bas";

// affiche en du text en justifier dans les x et centr dans les y
textJustifyCentre.Setup(fontTexture,ft_texture);
textJustifyCentre.SetClipping((100,100);
textJustifyCentre.SetAlign(align_justify,align_center);
textJustifyCentre.value="Justify centr";
textJustifyCentre.SetWrapping(FALSE,FALSE,FALSE); // affiche une seul ligne
- ou -
textJustifyCentre.SetWrapping(TRUE,FALSE,FALSE); // revient a la ligne si la ligne est pleine
- ou -
textJustifyCentre.SetWrapping(FALSE,TRUE,FALSE); // affiche une seul ligne finis par un mots
- ou -
textJustifyCentre.SetWrapping(TRUE,TRUE,FALSE); // retour a la ligne si cela coup un mots de fin de ligne
\endcode
*/

class Text3D : public BasicText3D
{
private:
	BOOL enableClipping;		//!< clipping activ ?
	Point2D<GLdouble> clipping;	//!< delimiteur de taille
	Point2D<GLdouble> realClipping;	//!< delimiteur de taille une fois pris en compt le scal
	int idCallList;				//!< numero de la list pour optimiser les perfs
	BOOL isChanged;				//!< y a t'il eut des changements
	GLdouble yJustify;			//!< taille de la justification dans les Y

	MyString value;
	MyList<ParsingLine> lines;			//!< lignes une fois pars
	MyList<ParsingWord> words;			//!< mots
	MyList<ParsingLine> layoutLines;	//!< lignes a ajouter pour la mise en page
	MyList<ParsingWord> layoutWords;	//!< mots a ajouter pour la mise en page
	MyList<ParsingLine *> linesUses;	//!< lignes retenu

	void DrawLine(ParsingLine *line);
	float AddLine(ParsingLine *line);

public:
	TextAlign xAlign;			//!< type de justification dans les X
	TextAlign yAlign;			//!< type de justification dans les Y
	GLdouble xJustifyLimit;		//!< limit de la taille max de justification dans les X, pour eviter des justications trop importente
	GLdouble yJustifyLimit;		//!< limit de la taille max de justification dans les Y, pour eviter des justications trop importente
	BOOL wordWrap;				//!< un mots est il secable ?
	BOOL lineWrap;				//!< une ligne est elle secable ?
	BOOL letterWrap;			//!< une lettre est elle secable ?

	Text3D();
	~Text3D();
	void operator=(Text3D &src);
	void SetJustifyLimit(GLdouble x,GLdouble y);
	void SetWrapping(BOOL line,BOOL word,BOOL letter);
	void SetClipping(double x,double y);
	void SetClipping(Point2D<double> newClipping);
	void SetAlign(TextAlign x=align_center,TextAlign y=align_center);
	void Setup(FTFont *f,FontType t=ft_texture);
	BOOL Draw();
	BOOL DrawPicking();
	void ComputeSize(MyList<ParsingLine> *linesCur,BOOL isWordSizeCompute);
	void ComputeLineSize(ParsingLine *line,BOOL isWordSizeCompute);

	float ComputeLayout();
	void ApplyRealClipping();

	void SetText(char *text);
	char * GetText();
	BOOL AnalyseLex(int indBegin,int *indEnd,BOOL *isEnd,BOOL *isReturn);
	void AnalyseGram();
};

#endif